﻿namespace Raytracer

    open Raytracer
    open Raytracer.Math
    open Raytracer.Primitives

    type Camera(pos : Vector3, lookAt : Vector3, sx : float, sy : float) = 
       let fwd  = lookAt - pos
       
       let down = new Vector3(0.0, 1.0, 0.0)
       let right = sx * (fwd .** down).Unit
       let up    = sy * (fwd .** right).Unit
       
       member this.Voxel(x, y) = pos + fwd + (2.0 * x - 1.0) * right + (2.0 * y - 1.0) * up
       member this.CameraPos = pos

    type Light(position : Vector3, color : Color) =
        member this.Position = position
        member this.Color    = color

    type Scene(background : Color,
               camera     : Camera,
               lights     : Light seq,
               objects    : SceneObject seq) =

        member this.Objects    = objects
        member this.Lights     = lights
        member this.Camera     = camera
        member this.Background = background

    [<AbstractClass>]
    type Raytracer(scene          : Scene,
                   resolution     : System.Drawing.Size,
                   recursiveDepth : int) =
                   
        let voxel(x, y) = scene.Camera.Voxel(float x / float resolution.Width, float y / float resolution.Height)
        let screenPixels = seq { for x = 0 to resolution.Width - 1 do for y = 0 to resolution.Height - 1 do yield (x, y) }
        
        // Ersten Schnittpunkt des Sehstrahls mit den Objekten der Szene berechnen
        let firstIntersection ray =
           let intersections = seq { for obj in scene.Objects do 
                                         let t = obj.IntersectRay ray
                                         if t >= 0.0 then yield (t, obj) }
                                          
           if intersections |> Seq.isEmpty
               then None
               else Some(intersections |> Seq.minBy fst)
        
        // Komplette Strahlverfolgung eines Sehstrahls von der Kamera zum Bildschirmpunkt (x, y)
        member this.Raytrace(x, y) =
            let ray = Ray.FromPoints(scene.Camera.CameraPos, voxel(x, y))
            
            // Rekursives Strahlverfolgen (Strahl, Verbleibende Rekursionstiefe)
            let rec raytraceRec ray depth =
                match firstIntersection ray with
                | None         -> scene.Background
                | Some(t, obj) -> 
                    let pos = ray.At t
                    let surface = obj.Surface
                    let norm = obj.Normal pos
                    
                    // Reflektierten Strahl vorberechnen (Reflexionsgesetz)
                    let d = ray.Direction 
                    let reflectedDir = (d - (2.0 * norm .* d) * norm).Unit
                    
                    // Erzeugte Farben aller Lichtquellen aufsummieren
                    let lightSum = 
                        scene.Lights |> Seq.sumBy (fun light ->
                             
                             // Strahl zur aktuellen Lichtquelle nach Schatten werfenden Hindernissen absuchen
                             let lightRay = Ray.FromPoints(light.Position, pos)
                             
                             // Abfragefunktion bereitstellen, ob einer Schnittstelle im Schatten liegt
                             let (|Illuminated|Shaded|) = function
                                 | None -> Illuminated
                                 | Some(_, x) when x = obj -> Illuminated
                                 | _ -> Shaded
                             
                             match firstIntersection lightRay with
                             | Shaded -> Color.Zero
                             | Illuminated -> 
                                 let phi      = -lightRay.Direction .* norm
                                 let theta    = -lightRay.Direction .* reflectedDir
                                 let direct   = if phi   >= 0.0 then light.Color * phi else Color.Zero
                                 let specular = if theta >= 0.0 then light.Color * (theta ** (surface.Roughness pos)) else Color.Zero 
                                 
                                 (direct .* surface.Diffuse pos) + (specular .* surface.Specular pos)
                        )     
                    
                    // Wenn eine reflektierende Oberfläche getroffen wurde, rekursiv den reflektierten Strahl weiterverfolgen
                    if surface.Reflection pos <= 0.0 || depth <= 0
                        then lightSum + new Color(0.5, 0.5, 0.5)
                        else 
                            let reflectedRay = new Ray(pos + 0.0001 * reflectedDir, reflectedDir)
                            lightSum + surface.Reflection pos * raytraceRec reflectedRay (depth - 1) 
                                                    
            raytraceRec ray recursiveDepth
        
        abstract member Reset: unit -> unit
        abstract member SetPixel : int * int * SysColor -> unit

        member this.Render() = 
            this.Reset()
            screenPixels |> Seq.toArray |> Array.Parallel.iter (fun (x, y) -> this.SetPixel(x, y, this.Raytrace(x, y).ToColor))